/*
 * Copyright 2012-2020 Open Source Security, Inc.
 *
 * This file is part of paxctld.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/inotify.h>
#include <sys/xattr.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <syslog.h>
#include <limits.h>
#include <signal.h>
#include <dirent.h>

#define PAXCTLD_VERSION "1.2.5"

#define DEFAULT_PAXCTLD_CONF "/etc/paxctld.conf"
#define DEFAULT_PERUSER_PAXCTLD_CONF ".paxctld/paxctld.conf"

#define DEFAULT_PAXCTLD_CONF_DIR "/etc/paxctld.d"
#define DEFAULT_PERUSER_PAXCTLD_CONF_DIR ".paxctld/paxctld.d"

#define MAX_CONFIG_ENTRIES 16384

#define INOTIFY_FLAGS (IN_DONT_FOLLOW | IN_ATTRIB | IN_CREATE | IN_DELETE_SELF | IN_MOVE_SELF | IN_MOVED_TO)
#define DIR_INOTIFY_FLAGS (INOTIFY_FLAGS &~ IN_ATTRIB)


#define PAX_PAGEEXEC_ON  0x00000001
#define PAX_SEGMEXEC_ON  0x00000002
#define PAX_MPROTECT_ON  0x00000004
#define PAX_ASLR_ON      0x00000008
#define PAX_EMUTRAMP_ON  0x00000010

#define PAX_XATTR_FLAGS	"user.pax.flags"

#define PAX_DEFAULT_FLAGS	\
	(PAX_PAGEEXEC_ON|PAX_SEGMEXEC_ON|PAX_MPROTECT_ON|PAX_ASLR_ON)

enum watch_type { WATCH_FILE, WATCH_DIR };

struct conf_entry {
	char *requested_path;
	char *existing_path;
	enum watch_type watch_type;
	unsigned int pax_flags;
	int watch_id;
	int nonroot;
};

struct paxctld_config {
	struct conf_entry *entries;
	unsigned int count;
};

static struct paxctld_config config;
static int ino = -1;

static int do_daemonize;
static char *pidfile;

static int root_mode;

static int quiet;
#define gr_syslog(level, ...) do {				\
	if (!quiet) {						\
		if (do_daemonize)				\
			syslog(level, ## __VA_ARGS__);		\
		else						\
			fprintf(stderr, ## __VA_ARGS__);	\
	}							\
} while (0)

static char *gr_strdup(const char *str)
{
	char *ret = strdup(str);
	if (ret == NULL) {
		fprintf(stderr, "Unable to allocate memory.\n");
		exit(EXIT_FAILURE);
	}	return ret;
}

static size_t decode_pax_flags(unsigned int flags, char *buf)
{
	buf[0] = '\0';
	if (!(flags & PAX_EMUTRAMP_ON))
		strcat(buf, "e");
	if (!(flags & PAX_PAGEEXEC_ON))
		strcat(buf, "p");
	if (!(flags & PAX_SEGMEXEC_ON))
		strcat(buf, "s");
	if (!(flags & PAX_MPROTECT_ON))
		strcat(buf, "m");
	if (!(flags & PAX_ASLR_ON))
		strcat(buf, "r");
	return strlen(buf);
}

static unsigned int encode_pax_flags(const char *conf)
{
	unsigned int ret = PAX_DEFAULT_FLAGS;
	const char *p = conf;
	while (*p) {
		switch (*p) {
		case 'E':
			ret |= PAX_EMUTRAMP_ON;
			break;
		case 'p':
			ret &= ~PAX_PAGEEXEC_ON;
			break;
		case 'm':
			ret &= ~PAX_MPROTECT_ON;
			break;
		case 'r':
			ret &= ~PAX_ASLR_ON;
			break;
		case 's':
			ret &= ~PAX_SEGMEXEC_ON;
			break;
		default:
			fprintf(stderr, "Unknown character: \"%c\" in PaX configuration string: \"%s\".  Permitted characters are: \"pEmrs\".\n", *p, conf);
			exit(EXIT_FAILURE);
		}
		p++;
	};

	return ret;
}

static char *get_parent_dir(char *path)
{
	char *tmpp = strrchr(path, '/');
	if (tmpp) {
		if (tmpp == path)
			tmpp[1] = '\0';
		else
			tmpp[0] = '\0';
	}

	return path;
}

static void init_watches(void);

static int add_watch(int inotify, const char *pathname, unsigned int flags)
{
	int ret = inotify_add_watch(inotify, pathname, flags);
	if (ret == -1 && errno == ENOSPC) {
		init_watches();
		ret = inotify_add_watch(inotify, pathname, flags);
	}
	return ret;
}

static void add_watch_to_closest_path(struct conf_entry *entry)
{
	char *path = gr_strdup(entry->requested_path);
	enum watch_type type = WATCH_FILE;
	int id;

	id = add_watch(ino, path, INOTIFY_FLAGS);
	while (id == -1 && errno == ENOENT && strcmp(path, "/")) {
		// keep stripping path components until we reach an existing directory
		get_parent_dir(path);
		id = add_watch(ino, path, DIR_INOTIFY_FLAGS);
		type = WATCH_DIR;
	}

	// drop possible old watch
	if (entry->watch_id >= 0)
		inotify_rm_watch(ino, entry->watch_id);

	entry->watch_id = id;

	// update path we're watching
	if (entry->existing_path)
		free(entry->existing_path);

	entry->existing_path = path;
	entry->watch_type = type;
}

/* do safe setting of user xattrs */
static int set_xattr(struct conf_entry *entry)
{
	struct stat st;
	uid_t linkuid;
	char path[PATH_MAX+1];
	char pax_flags_str[16];
	char old_flags[16] = { };
	ssize_t len;
	int ret;

	if (lstat(entry->requested_path, &st)) {
		if (errno == ENOENT)
			return 0;

		goto error;
	}

	linkuid = st.st_uid;
	if (root_mode && !entry->nonroot && linkuid)
		goto error2;

	if (!realpath(entry->requested_path, path))
		return 0;

	if (lstat(path, &st)) {
		if (errno == ENOENT)
			return 0;

		goto error;
	}

	if ((entry->nonroot || !root_mode) && st.st_uid != linkuid && linkuid)
		goto error2;

	if (!S_ISREG(st.st_mode))
		goto error3;

	len = decode_pax_flags(entry->pax_flags, pax_flags_str);

	// don't do anything if the flags are already set
	if (lgetxattr(path, PAX_XATTR_FLAGS, old_flags, sizeof(old_flags)-1) == len &&
	    strcmp(pax_flags_str, old_flags) == 0)
		return 0;

	ret = lsetxattr(path, PAX_XATTR_FLAGS, pax_flags_str, strlen(pax_flags_str), 0);
	if (ret == -1) {
		if (errno == ENOENT)
			return 0;
		goto error;
	} else {
		gr_syslog(LOG_INFO, "Restored PaX flags on \"%s\".\n", entry->existing_path);
	}

	return 1;
error:
	gr_syslog(LOG_ERR, "Unable to set extended attribute on \"%s\".  Error: %s\n", entry->requested_path, strerror(errno));
	exit(EXIT_FAILURE);
error2:
	gr_syslog(LOG_ERR, "Unable to set extended attribute on \"%s\".  Error: owner of symlink did not match that of target.\n", entry->requested_path);
	exit(EXIT_FAILURE);
error3:
	gr_syslog(LOG_ERR, "Declining to set PaX flags on \"%s\".  Error: not a regular file.\n", entry->requested_path);
	exit(EXIT_FAILURE);
}

static void init_watches(void)
{
	unsigned int i;

	if (ino >= 0)
		close(ino);

	ino = inotify_init();
	if (ino < 0) {
		gr_syslog(LOG_ERR, "Fatal: Unable to initialize inotify system: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	for (i = 0; i < config.count; i++) {
		struct conf_entry *entry = &config.entries[i];

		add_watch_to_closest_path(entry);
		/* mark the binary if possible */
		set_xattr(entry);
	}
}

static const char quoted_scan_string[] = "\"%4095[^\"]\" %15s %9s";
static const char unquoted_scan_string[] = "%4095s %15s %9s";

static void parse_config(const char *confpath, struct paxctld_config *config)
{
	FILE *f = fopen(confpath, "r");
	char buf[8192] = { 0 };
	char *p;
	char path[4096] = { 0 };
	char flags[16] = { 0 };
	char rootopt[10] = { 0 };
	unsigned long lineno = 0;
	int nonroot;
	int ret;

	const char *scan_string;

	if (f == NULL) {
		fprintf(stderr, "Unable to open configuration file: %s\nError: %s\n", confpath, strerror(errno));
		exit(EXIT_FAILURE);
	}

	while(fgets(buf, sizeof(buf) - 1, f)) {
		lineno++;

		p = buf;
		while (*p == ' ' || *p == '\t')
			p++;
		// ignore comment and empty lines
		if (*p == '#' || *p == '\n')
			continue;

		nonroot = 0;
		// if the path is quoted (i.e. has spaces), accomodate that
		if (*p == '"')
			scan_string = quoted_scan_string;
		else
			scan_string = unquoted_scan_string;

		ret = sscanf(p, scan_string, path, flags, rootopt);
		if (ret == 3) {
			if (!strcmp(rootopt, "nonroot"))
				nonroot = 1;
			else
				ret = -1;
		}

		if (ret < 2) {
			fprintf(stderr, "Invalid configuration on line %lu of %s.\nSyntax is: </absolute/path> <PaX flags> [nonroot]\n", lineno, confpath);
			exit(EXIT_FAILURE);
		}

		if (config->count >= MAX_CONFIG_ENTRIES) {
			fprintf(stderr, "Exceeded maximum number of config entries.\n");
			exit(EXIT_FAILURE);
		}

		config->entries[config->count].watch_id = -1;
		config->entries[config->count].nonroot = nonroot;
		config->entries[config->count].requested_path = gr_strdup(path);
		config->entries[config->count].pax_flags = encode_pax_flags(flags);

		config->count++;
	}

	fclose(f);
}

static void usage(const char *name)
{
	fprintf(stderr, "Usage: %s [-c config_file] [-d] [-p pid_file] [-q] [-v]\n", name);
	exit(EXIT_FAILURE);
}

static void handle_event(struct inotify_event *event, struct conf_entry *confentry)
{
	if (confentry->watch_type == WATCH_DIR) {
		// try to refine the watched path
		add_watch_to_closest_path(confentry);
		if (confentry->watch_type == WATCH_FILE)
			gr_syslog(LOG_INFO, "File %s created.\n", confentry->existing_path);
	} else {
		// might be deleted, adapt watched path
		if (event->mask & (IN_DELETE_SELF | IN_MOVE_SELF))
			add_watch_to_closest_path(confentry);
		if (confentry->watch_type == WATCH_DIR)
			gr_syslog(LOG_INFO, "File %s deleted.\n", confentry->existing_path);
	}

	/* if after processing the existing file matches the requested file, then set
	   extended attributes on the file */
	if (confentry->watch_type == WATCH_FILE)
		set_xattr(confentry);
}

static void daemonize(void)
{
	pid_t pid;

	pid = fork();
	if (pid == 0) {
		if (setsid() < 0)
			exit(EXIT_FAILURE);
		signal(SIGCHLD, SIG_IGN);
		signal(SIGHUP, SIG_IGN);
		pid = fork();
		if (pid == 0) {
			FILE *f;
			int i;
			if (chdir("/"))
				exit(EXIT_FAILURE);
			for (i = 0; i <= sysconf(_SC_OPEN_MAX); i++)
				if (i != ino)
					close(i);
			if (pidfile) {
				f = fopen(pidfile, "w");
				fprintf(f, "%u\n", getpid());
				fclose(f);
			}

			openlog("paxctld", 0, LOG_DAEMON);

			return;
		} else
			exit(EXIT_SUCCESS);
	} else
		exit(EXIT_SUCCESS);
}

static char *make_user_config_path(const char *user_path)
{
	char *config_path;

	config_path = calloc(1, PATH_MAX);
	if (config_path == NULL) {
		fprintf(stderr, "Unable to allocate memory.\n");
		exit(EXIT_FAILURE);
	}
	snprintf(config_path, PATH_MAX-1, "%s/%s", getenv("HOME"), user_path);

	return config_path;
}

static const char *get_default_config_path(void)
{
	if (root_mode)
		return DEFAULT_PAXCTLD_CONF;

	return make_user_config_path(DEFAULT_PERUSER_PAXCTLD_CONF);
}

static const char *get_default_config_dir_path(void)
{
	if (root_mode)
		return DEFAULT_PAXCTLD_CONF_DIR;

	return make_user_config_path(DEFAULT_PERUSER_PAXCTLD_CONF_DIR);
}

int main(int argc, char *argv[])
{
	static const size_t buf_size = sizeof(struct inotify_event) + 4096;
	const char *config_path = NULL;
	const char *config_dir_path;
	unsigned int i;
	char flags[16] = { 0 };
	void *buf;
	DIR *dir;
	int opt;

	if (argc < 1)
		usage("paxctld");

	root_mode = !getuid();

	while ((opt = getopt(argc, argv, "vc:dp:q")) != -1) {
		switch (opt) {
		case 'v':
			printf("paxctld v%s\n", PAXCTLD_VERSION);
			printf("Copyright (C) 2012-2020 Open Source Security, Inc.\n");
			printf("License GPLv2: GNU GPL version 2 <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>\n");
			printf("This is free software: you are free to change and redistribute it.\n");
			printf("There is NO WARRANTY, to the extent permitted by law.\n");
			exit(0);
			break;
		case 'c':
			config_path = gr_strdup(optarg);
			break;
		case 'd':
			do_daemonize = 1;
			break;
		case 'p':
			pidfile = gr_strdup(optarg);
			break;
		case 'q':
			quiet = 1;
			break;
		default:
			fprintf(stderr, "Unknown option: \"%c\".", opt);
			usage(argv[0]);
		}
	}

	if (config_path == NULL)
		config_path = get_default_config_path();

	if (getxattr("/proc/self/exe", PAX_XATTR_FLAGS, flags, sizeof(flags)-1) == -1 && errno == ENOTSUP) {
		fprintf(stderr, "Fatal: Filesystem extended attribute support is not enabled on the current running kernel.\n");
		exit(EXIT_FAILURE);
	}

	config.count = 0;
	config.entries = calloc(MAX_CONFIG_ENTRIES, sizeof(struct conf_entry));
	if (config.entries == NULL) {
		fprintf(stderr, "Unable to allocate memory.\n");
		exit(EXIT_FAILURE);
	}

	parse_config(config_path, &config);

	config_dir_path = get_default_config_dir_path();
	dir = opendir(config_dir_path);
	if (dir) {
		struct dirent *conf_file;
		char tmppath[PATH_MAX] = { 0 };

		while ((conf_file = readdir(dir))) {
			struct stat st;
			if (conf_file->d_name[0] == '.')
				continue;
			snprintf(tmppath, sizeof(tmppath)-1, "%s/%s", config_dir_path, conf_file->d_name);
			if (stat(tmppath, &st))
				continue;
			if (!S_ISREG(st.st_mode))
				continue;

			parse_config(tmppath, &config);
		}
		closedir(dir);
	}

	buf = calloc(1, buf_size);
	if (buf == NULL) {
		fprintf(stderr, "Fatal: Unable to allocate memory.\n");
		exit(EXIT_FAILURE);
	}

	if (do_daemonize)
		daemonize();

	gr_syslog(LOG_INFO, "paxctld initialized.\n");

	init_watches();

	for (;;) {
		struct inotify_event *event;
		ssize_t cnt;
		int done;
		void *e;

		cnt = read(ino, buf, buf_size);
		if (cnt < 0) {
			gr_syslog(LOG_ERR, "Failed to get inotify events: %s\n", strerror(errno));
			break;
		}

		// process all events
		done = 0;
		for (e = buf; !done && e < buf + cnt; e += sizeof(event) + event->len) {
			event = e;

			// if it's a delete, we need to remove the watch on the file and add it to its parent directory
			for (i = 0; i < config.count; i++) {
				struct conf_entry *entry = &config.entries[i];

				// event queue overflow? start all over again
				if (event->wd == -1) {
					init_watches();
					done = 1;
					break;
				}

				if (event->wd != entry->watch_id)
					continue;

				handle_event(event, entry);
			}
		}
	}

	gr_syslog(LOG_INFO, "Terminating\n");

	return 0;
}
